@@ -13,26 +13,22 @@ use constant {
     _RESOLUTION_CLASS => 'Promise::ES6::_RESOLUTION',
     _REJECTION_CLASS  => 'Promise::ES6::_REJECTION',
     _PENDING_CLASS    => 'Promise::ES6::_PENDING',
-
-    _DEBUG => 0,
 };
 
 use constant {
-    _PROMISE_ID_IDX  => 0,
-    _PID_IDX         => _DEBUG + 0,
-    _CHILDREN_IDX    => _DEBUG + 1,
-    _VALUE_SR_IDX    => _DEBUG + 2,
-    _DETECT_LEAK_IDX => _DEBUG + 3,
-    _ON_RESOLVE_IDX  => _DEBUG + 4,
-    _ON_REJECT_IDX   => _DEBUG + 5,
-    _IS_FINALLY_IDX  => _DEBUG + 6,
+    _PID_IDX         => 0,
+    _CHILDREN_IDX    => 1,
+    _VALUE_SR_IDX    => 2,
+    _DETECT_LEAK_IDX => 3,
+    _ON_RESOLVE_IDX  => 4,
+    _ON_REJECT_IDX   => 5,
+    _IS_FINALLY_IDX  => 6,
 };
 
 # "$value_sr" => $value_sr
 our %_UNHANDLED_REJECTIONS;
 
-my $_debug_promise_id = 0;
-sub _create_promise_id { return $_debug_promise_id++ . "-$_[0]" }
+use constant _THEN => 'then';
 
 sub new {
     my ( $class, $cr ) = @_;
@@ -45,15 +41,12 @@ sub new {
     my @children;
 
     my $self = bless [
-        ( _DEBUG ? undef : () ),
         $$,
         \@children,
         $value_sr,
         $Promise::ES6::DETECT_MEMORY_LEAKS,
     ], $class;
 
-    $self->[_PROMISE_ID_IDX] = _create_promise_id($self) if _DEBUG;
-
     # NB: These MUST NOT refer to $self, or else we can get memory leaks
     # depending on how $resolver and $rejector are used.
     my $resolver = sub {
@@ -63,7 +56,7 @@ sub new {
         # NB: UNIVERSAL::can() is used in order to avoid an eval {}.
         # It is acknowledged that many Perl experts strongly discourage
         # use of this technique.
-        if ( UNIVERSAL::can( $$value_sr, 'then' ) ) {
+        if ( UNIVERSAL::can( $$value_sr, _THEN ) ) {
             _repromise( $value_sr, \@children, $value_sr );
         }
         elsif (@children) {
@@ -96,36 +89,51 @@ sub new {
 }
 
 sub then {
-    return $_[0]->_then_or_finally(@_[1, 2]);
-}
-
-sub finally {
-
-    # There’s no reason to call finally() without a callback
-    # since it would just be a no-op.
-    die 'finally() requires a callback!' if !$_[1];
-
-    return $_[0]->_then_or_finally($_[1], undef, 1);
-}
-
-sub _then_or_finally {
-    my ($self, $on_resolve_or_finish, $on_reject, $is_finally) = @_;
+    my ( $self, $on_resolve, $on_reject ) = @_;
 
     my $value_sr = bless( \do { my $v }, _PENDING_CLASS() );
 
     my $new = bless [
-        ( _DEBUG ? undef : () ),
         $$,
         [],
         $value_sr,
         $Promise::ES6::DETECT_MEMORY_LEAKS,
-        $on_resolve_or_finish,
+        $on_resolve,
         $on_reject,
-        $is_finally,
       ],
       ref($self);
 
-    $new->[_PROMISE_ID_IDX] = _create_promise_id($new) if _DEBUG;
+    if ( _PENDING_CLASS eq ref $self->[_VALUE_SR_IDX] ) {
+        push @{ $self->[_CHILDREN_IDX] }, $new;
+    }
+    else {
+
+        # $self might already be settled, in which case we immediately
+        # settle the $new promise as well.
+
+        $new->_settle( $self->[_VALUE_SR_IDX] );
+    }
+
+    return $new;
+}
+
+sub finally {
+    my ( $self, $on_finish ) = @_;
+
+    my $value_sr = bless( \do { my $v }, _PENDING_CLASS() );
+
+    my $new = bless(
+        [
+            $$,
+            [],
+            $value_sr,
+            $Promise::ES6::DETECT_MEMORY_LEAKS,
+            $on_finish,
+            undef,
+            1,  # is finally
+        ],
+        ref($self),
+    );
 
     if ( _PENDING_CLASS eq ref $self->[_VALUE_SR_IDX] ) {
         push @{ $self->[_CHILDREN_IDX] }, $new;
@@ -187,8 +195,6 @@ sub _settle {
 
     @{$self}[ _ON_RESOLVE_IDX, _ON_REJECT_IDX ] = ();
 
-    # In some contexts this function runs quite a lot,
-    # so caching the is-promise lookup is useful.
     my $value_sr_contents_is_promise = 1;
 
     if ($callback) {
@@ -209,7 +215,7 @@ sub _settle {
             # just run a successful “catch” block, so resolution is correct.
 
             # If $new_value IS a promise, though, then we have to wait.
-            if ( !UNIVERSAL::can( $new_value, 'then' ) ) {
+            if ( !UNIVERSAL::can( $new_value, _THEN ) ) {
                 $value_sr_contents_is_promise = 0;
 
                 if ($self_is_finally) {
@@ -251,11 +257,15 @@ sub _settle {
         # indicates # (i.e., resolution or rejection) is now $self’s state
         # as well.
 
-        # NB: We should NEVER be here if the promise is from finally().
+        if ($self_is_finally) {
+            $self->[_VALUE_SR_IDX] = $final_value_sr;
+        }
+        else {
+            bless $self->[_VALUE_SR_IDX], ref($final_value_sr);
+            ${ $self->[_VALUE_SR_IDX] } = $$final_value_sr;
+        }
 
-        bless $self->[_VALUE_SR_IDX], ref($final_value_sr);
-        ${ $self->[_VALUE_SR_IDX] } = $$final_value_sr;
-        $value_sr_contents_is_promise = UNIVERSAL::can( $$final_value_sr, 'then' );
+        $value_sr_contents_is_promise = UNIVERSAL::can( $$final_value_sr, _THEN );
 
         if ($settle_is_rejection) {
             $_UNHANDLED_REJECTIONS{ $self->[_VALUE_SR_IDX] } = $self->[_VALUE_SR_IDX];
